import os
import logging
from enum import Enum
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from COMPS import Client
from COMPS.Data import OutputFileMetadata

logger = logging.getLogger('COMPS.Data.AssetManager')

max_asset_zip_file_size = 1073741824


def retrieve_asset_files(asset_files, as_zip=False):
    if as_zip:
        total_zip_size = 0
        retrieve_assets = []

        for af in asset_files:
            total_zip_size += af.length

        if total_zip_size > max_asset_zip_file_size:
            raise RuntimeError('Total size of requested files ({0}) exceeds maximum allowed ({1})' \
                               .format(total_zip_size, max_asset_zip_file_size))

        zip_md = _do_asset_zip_url_request(asset_files)

        logger.debug('adding zip download url = ' + zip_md.url)
        retrieve_assets.append(zip_md)
    else:
        retrieve_assets = asset_files

    byte_arrs = _retrieve_files_from_metadata(retrieve_assets, lambda x: x.url if as_zip else x.uri)

    return byte_arrs[0] if as_zip else byte_arrs

def retrieve_output_file_info(entity_type, entity_id, paths, job=None):
    query_params = {'flatten': 1}
    if job:
        if entity_type == EntityType.WorkItems:
            raise RuntimeError('Setting \'job\' parameter for WorkItem is not valid')
        query_params['hpcjobid'] = job._id

    req_path = "/asset/{0}/{1}/Output".format(entity_type.name, str(entity_id))
    resp = Client.get(req_path,
                      params=query_params)

    json_resp = resp.json()

    if 'Resources' not in json_resp:
        raise RuntimeError('Malformed Asset Service response!')

    ofmd_map = {}

    for ofmd_json in json_resp['Resources']:
        ofmd_json = OutputFileMetadata.rest2py(ofmd_json)

        # logger.debug('Asset:')
        # logger.debug(json.dumps(ofmd_json, indent=4))

        ofmd = OutputFileMetadata.__internal_factory__(**ofmd_json)

        file_path = ofmd.friendly_name if ofmd.path_from_root == '.' else ofmd.path_from_root + '/' + ofmd.friendly_name
        logger.debug('adding [{0} -> {1}] to map'.format(file_path, ofmd.friendly_name))
        ofmd_map[file_path.lower()] = ofmd

    if paths is None or len(paths) == 0:    # paths not specified, return everything
        return ofmd_map.values()

    file_info = []

    for p in paths:
        ofmd = ofmd_map.get(p.lower().replace('\\', '/'))

        if not ofmd:
            raise RuntimeError('Couldn\'t find file for path \'{0}\''.format(p))

        file_info.append(ofmd)

    return file_info

def retrieve_output_files_from_info(entity_type, entity_id, metadata, job=None, as_zip=False):
    if as_zip:
        if job and entity_type == EntityType.WorkItems:
            raise RuntimeError('Setting \'job\' parameter for WorkItem is not valid')

        total_zip_size = 0
        retrieve_files = []
        ids = []

        for ofmd in metadata:
            total_zip_size += ofmd.length
            ids.append(ofmd._id)

        if total_zip_size > max_asset_zip_file_size:
            raise RuntimeError('Total size of requested files ({0}) exceeds maximum allowed ({1})' \
                               .format(total_zip_size, max_asset_zip_file_size))

        zip_md = _do_output_zip_url_request(entity_type, entity_id, ids, job)

        logger.debug('adding zip download url = ' + zip_md.url)
        retrieve_files.append(zip_md)
    else:
        retrieve_files = metadata

    byte_arrs = _retrieve_files_from_metadata(retrieve_files, lambda x: x.url)

    return byte_arrs[0] if as_zip else byte_arrs

def _retrieve_files_from_metadata(files_metadata, get_url_fn):
    byte_arrs = []

    for fmd in files_metadata:
        url = get_url_fn(fmd)
        i = url.find('/asset/')
        if i == -1:
            raise RuntimeError('Unable to parse asset url: ' + url)

        resp = Client.get(url[i:])  # tried "stream=True", but ran into issues with ZipFile.  Just do this for now...

        # stream = resp.raw

        byte_arrs.append(resp.content)

    return byte_arrs

def _do_output_zip_url_request(entity_type, entity_id, ids, job):
    query_params = {'zip': 1}
    if job:
        query_params['hpcjobid'] = job._id

    req_path = "/asset/{0}/{1}/Output".format(entity_type.name, str(entity_id))
    resp = Client.post(req_path,
                       json={'ContainerType': 'CompressedZip', 'Ids': [ str(ofid) for ofid in ids ]},
                       params=query_params)

    json_resp = resp.json()

    if 'Resources' not in json_resp or len(json_resp['Resources']) != 1:
        raise RuntimeError('Malformed Asset Service response!')

    ofmd_json = OutputFileMetadata.rest2py(json_resp['Resources'][0])

    # logger.debug('Asset:')
    # logger.debug(json.dumps(ofmd_json, indent=4))

    ofmd = OutputFileMetadata.__internal_factory__(**ofmd_json)

    return ofmd

def _do_asset_zip_url_request(asset_files):
    # query_params = {'zip': 1}

    entries = [ {
                    'FileNameForAsset': _combine_asset_path_and_file(af),
                    'MD5': str(af.md5_checksum)
                } for af in asset_files ]

    resp = Client.post("/ZipAssets",
                       json={'ContainerType': 'CompressedZip', 'Assets': entries}) #,
                       # params=query_params)

    json_resp = resp.json()

    if 'Resources' not in json_resp or len(json_resp['Resources']) != 1:
        raise RuntimeError('Malformed Asset Service response!')

    ofmd_json = OutputFileMetadata.rest2py(json_resp['Resources'][0])

    # logger.debug('Asset:')
    # logger.debug(json.dumps(ofmd_json, indent=4))

    ofmd = OutputFileMetadata.__internal_factory__(**ofmd_json)

    return ofmd

def _combine_asset_path_and_file(asset):
    if asset.relative_path is None or asset.relative_path == '':
        return asset.file_name
    else:
        return os.path.join(asset.relative_path, asset.file_name).replace('/','\\')

class EntityType(Enum):
    Simulations = 0
    WorkItems = 1
